#ifdef __cplusplus

#include <algorithm>
#include <functional>
#include <memory>
#include <string>
#include <type_traits>
#include <utility>
#include <vector>

#include <QCborValue>
#include <QDebug>
#include <QList>
#include <QMetaObject>
#include <QMetaType>
#include <QObject>
#include <QPointer>
#include <QString>
#include <QTextFormat>
#include <QVariant>
#include <QWidget>
#include <QtGlobal>

#endif
