<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE gui SYSTEM "kpartgui.dtd">
<gui name="kategitblameplugin" library="kategitblameplugin" version="3" translationDomain="kategitblameplugin">
    <MenuBar>
        <Menu name="tools">
            <text>&amp;Tools</text>
            <Action name="git_blame_show" group="tools_git_blame"/>
            <Action name="git_blame_toggle" group="tools_git_blame"/>
        </Menu>
    </MenuBar>
</gui>
<!-- kate: space-indent on; indent-width 4; replace-tabs on; -->
