#pragma once

namespace TargetKeys
{
constexpr static QLatin1String F_TARGET("target");
constexpr static QLatin1String F_DEBUGGER("debuggerKey");
constexpr static QLatin1String F_PROFILE("debuggerProfile");
constexpr static QLatin1String F_FILE("file");
constexpr static QLatin1String F_WORKDIR("workdir");
constexpr static QLatin1String F_ARGS("args");
constexpr static QLatin1String F_PID("pid");
constexpr static QLatin1String F_IS_LAUNCH_JSON("loadedFromLaunchJson");
}
