<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE gui SYSTEM "kpartgui.dtd">
<gui name="keyboardmacrosplugin" library="libkeyboardmacrosplugin" version="2" translationDomain="katekeyboardmacros">
  <MenuBar>
    <Menu name="tools">
      <text>&amp;Tools</text>
      <Separator/>
      <Action name="keyboardmacros" />
    </Menu>
  </MenuBar>
  <ToolBar name="keyboardmacros">
    <text>Keyboard Macros</text>
  </ToolBar>
</gui>
<!-- kate: space-indent on; indent-width 4; replace-tabs on; -->
