<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE gui SYSTEM "kpartgui.dtd">
<gui name="katecloseexcept" library="katecloseexceptplugin" version="4" translationDomain="katecloseexceptplugin">
  <MenuBar>
    <Menu name="file">
      <text>&amp;File</text>
      <Action name="file_close_except" group="close_merge"/>
      <Action name="file_close_like" group="close_merge"/>
    </Menu>
  </MenuBar>
</gui>
<!-- kate: space-indent on; indent-width 4; replace-tabs on; -->
