<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE gui SYSTEM "kpartgui.dtd">
<gui name="kategdbplugin" library="kategdbplugin" version="5" translationDomain="kategdbplugin">
  <MenuBar>
    <Menu name="debug">
      <Text>&amp;Debug</Text>
      <Action name="targets"/>
      <Action name="argLists"/>
      <Separator/>
      <Action name="debug"/>
      <Action name="kill"/>
      <Action name="rerun"/>
      <Separator/>
      <Action name="toggle_breakpoint"/>
      <Action name="clear_all_breakpoints"/>
      <Action name="step_in"/>
      <Action name="step_over"/>
      <Action name="step_out"/>
      <Action name="move_pc"/>
      <Action name="run_to_cursor"/>
      <Action name="continue"/>
      <Separator/>
      <Action name="print_value"/>
      <Action name="debug_hot_reload"/>
      <Action name="debug_hot_restart"/>
      <Action name="debug_hot_reload_on_save"/>
    </Menu>
  </MenuBar>
  <Menu name="ktexteditor_popup" noMerge="1">
    <Action name="popup_gdb"/>
  </Menu>
  <ToolBar name="gdbplugin">
    <text>Debug Plugin</text>
    <Action name="toggle_breakpoint"/>
    <Action name="step_in"/>
    <Action name="step_over"/>
    <Action name="step_out"/>
    <Action name="run_to_cursor"/>
    <Action name="continue"/>
  </ToolBar>
</gui>
<!-- kate: space-indent on; indent-width 2; replace-tabs on; -->
