<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE gui SYSTEM "kpartgui.dtd">
<gui name="katesearch" library="libkatesearchplugin" version="7" translationDomain="katesearch">
    <MenuBar>
        <Menu name="edit">
            <text>&amp;Edit</text>
            <Action name="search_in_files"/>
            <Menu name="search_more">
                <text>Find in Files Variants</text>
                <Action name="go_to_next_match"/>
                <Action name="go_to_prev_match"/>
                <Action name="search_in_files_new_tab"/>
                <Action name="cut_searched_lines"/>
                <Action name="copy_searched_lines"/>
            </Menu>
        </Menu>
    </MenuBar>
</gui>
<!-- kate: space-indent on; indent-width 4; replace-tabs on; -->
